#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include "usb.h"
#include "usbfs.h"
#include "ether.h"

#include "mv8388.i"

static int mvreset(Ether *e);

typedef struct Libertas Libertas;
struct Libertas
{
	QLock l;
	Rendez vous;
	int cmdresp;
	void *cmd;
	int respsz;

	int macctl;
	int hwversion;
	int version;
	int region;
	int flags;

	char ssid[32];
	uchar bssid[Eaddrlen];
	int channel;

	char try_ssid[32];
	uchar try_bssid[Eaddrlen];
	int try_channel;

	int scanc;
	int scanlen;
	uchar scanb[512];
	char scans[1024];
};

#pragma	pack on

enum
{
	Mboot = 0x4c56524d,	/* MRVL in little endian */
	CBfwusb	= 0x01,
	CBfweeprom	= 0x02,
	CBupdateboot2	= 0x03,
	CBupdatefw	= 0x04,

	RBfail	= 0x00,
	RBok	= 0x01,
	RBunsupported	= 0x02,
};

struct bootcmd 
{
	uchar magic[4];
	uchar cmd;
	uchar result;
	uchar pad[10];	/* [2] for response */
};

static int
bootcmd(Ether *e, int cmd){
	struct bootcmd c;
	int n;

	memset(&c, 0, sizeof(c));
	PUT4(c.magic, Mboot);
	c.cmd = cmd;
	n = write(e->epout->dfd, &c, sizeof(c));
	if(n != sizeof(c))
		return -1;

	/* TODO: timeout see alarm(2) */
	n = read(e->epin->dfd, &c, sizeof(c));
	if(n < sizeof(c)-sizeof(c.pad))
		return -1;

	if(GET4(c.magic) != Mboot || c.cmd != cmd){
		dprint(2,"%s: wierd boot response: %.*H\n", argv0, n, &c);
		return -1;
	}
	return c.result;
}

struct fwpkt
{
	uchar cmd[4];
	uchar addr[4];
	uchar len[4];
	uchar crc[4];
	uchar seq[4];
	uchar data[512];
};

struct fwsync
{
	uchar crc[4];
	uchar seq[4];
};

static int
writefw(Ether *e, uchar *fw, int size)
{
	struct fwpkt pkt;
	struct fwsync s;
	uchar *p, *end;
	int len, seq, retry, n;

	seq=0;
	retry=0;

	end = fw+size;
	for(p = fw; p < end;){
		memcpy(&pkt, p, 16);
		p += 16;
		PUT4(pkt.seq, seq);	/* firmware doesn't include sequence numbers */
		len = GET4(pkt.len);
		ddprint(2, "%s: fw write cmd %ux addr %ux len %ux seq %ux offset %ulx\n", argv0, GET4(pkt.cmd), GET4(pkt.addr), len, seq, p - fw);
		assert(len <= sizeof(pkt.data));
		assert(end >= p + len);
		memcpy(pkt.data, p, len);
		p += len;
		seq++;
		n = write(e->epout->dfd, &pkt, 20 + len);
		if(n != 20 + len)
			return 0;

		n = read(e->epin->dfd, &s, sizeof(s));
		if(n < sizeof(s))
			return 0;
		ddprint(2, "%s: fw sync crc %x seq %x\n", argv0, GET4(s.crc), GET4(s.seq));
		if(GET4(s.crc) != 0){
			p -= 16 + len;
			seq--;
			if(++retry > 8)
				return 0;
		}
	}
	return 1;	
} 

/* TODO
 * evaluate possibility of loading firmware from eeprom
 * and using another program to flash firmware
 */
static int
mvfirmware(Ether *e)
{
	char tmp;
	dprint(2, "%s: initializing the MV8388\n", argv0);
	switch(bootcmd(e, CBfwusb)){
	case RBok:
		dprint(2, "%s: MV8388 loading firmware\n", argv0);
		writefw(e, mv8388fw, sizeof(mv8388fw));
		/* there are two welcome packets */
		read(e->epin->dfd, &tmp, 1);
		read(e->epin->dfd, &tmp, 1);
		return 1;
		break;
	case RBunsupported:
		dprint(2, "%s: MV8388 firmware already loaded\n", argv0);
		return 1;
		break;
	case RBfail:
	default:
		dprint(2, "%s: MV8388 fail!\n", argv0);
		return 0;
		break;
	}
}

/* after firmware load */
enum {
	Mdata = 0xbeadc0de,
	Mindication = 0xbeefface,
	Mrequest	= 0xf00dface,

	Cgetspec	= 0x03,
	Creset	= 0x05,
	Cscan	= 0x06,
	Cglog	= 0x0b,
	Cmacm	= 0x10,
	Cauth	= 0x11,
	Cwep	= 0x13,
	Csnmp	= 0x16,
	Cregmac	= 0x19,
	Cregbbp	= 0x1a,
	Cregrf	= 0x1b,
	Cradioctl	= 0x1c,
	Cchan	= 0x1d,
	Ctxpower	= 0x1e,
	Crssi	= 0x1f,
	Cantenna	= 0x20,
	Cpowsave	= 0x21,
	Cdatarate	= 0x22,
	Cdeauth	= 0x24,
	Cdisass	= 0x26,
	Cmacctl	= 0x28,
	Cahstart	= 0x2b,
	Cahjoin	= 0x2c,
	Cenrsn	= 0x2f,
	Cdeep	= 0x3e,
	Cahstop	= 0x40,
	Chostsl	= 0x43,
	Cwakeup	= 0x44,
	Cregion	= 0x4c,
	Cmac	= 0x4d,
	Cassociate	= 0x50,
	Cband	= 0x58,
	Ceeprom	= 0x59,
	Cgspi	= 0x5a,
	Cdomain	= 0x5b,
	Cwmm	= 0x5c,
	Ckey	= 0x5e,
	Ctpcinfo	= 0x5f,
	Ctpcadapt	= 0x60,
	Cchanann	= 0x61,
	Cmeasreq	= 0x62,
	Cmeasrep	= 0x63,
	Csleeppar	= 0x66,
	Csleepper	= 0x68,
	Cbcacfg	= 0x69,
	Cscancfg	= 0x6b,
	Cscanq	= 0x6c,
	Ccaldata	= 0x6d,
	Cwmmg	= 0x71,
	Ctpccfg	= 0x72,
	Cpacfg	= 0x73,
	Cfwwake	= 0x74,
	Csubscrib	= 0x75,
	Crateada	= 0x76,
	Ctxrateq	= 0x7f,
	Csetb2v	= 0xa5,
	Cbeaconc	= 0xb0,
	Cbeacons	= 0xcb,
	Csetmode	= 0xcc,

	Aget	= 0,
	Aset	= 1,
};

struct cmd {
	uchar magic[4];
	uchar cmd[2];
	uchar size[2];
	uchar sequence[2];
	uchar result[2];
	uchar data[];
};

/* mvreset may be used before everything is set up */
static int
mvreset(Ether *e)
{
	struct cmd cmd;
	PUT4(cmd.magic, Mrequest);
	PUT2(cmd.cmd, Creset);
	PUT2(cmd.size, sizeof(cmd)-4);
	PUT2(cmd.sequence, 255);
	PUT2(cmd.result, 0);
	return write(e->epout->dfd, &cmd, sizeof(cmd));
};

static int
mvcmd(Ether *e, int cmdn, int len, struct cmd *cmd, int size){
	Libertas *liber;
	int n;
	static ushort seq;

	liber = e->aux;
	qlock(&liber->l);
	seq++;
	liber->cmdresp = 0;
	liber->respsz = size;
	liber->cmd = cmd;

	PUT4(cmd->magic, Mrequest);
	PUT2(cmd->cmd, cmdn);
	PUT2(cmd->size, len);
	PUT2(cmd->sequence, seq);
	dprint(2, "%s: docmd: %d bytes: %.*H\n", argv0, len+4, len+4, cmd);
	n = write(e->epout->dfd, cmd, len+4);
	if(n != len+4){
		qunlock(&liber->l);
		liber->cmd = nil;
		return -1;
	};
	rsleep(&liber->vous);
	/* cmd now holds the response */
	n = liber->respsz;
	liber->cmd = nil;
	qunlock(&liber->l);
	dprint(2, "%s: docmd: %d bytes: %.*H\n", argv0, n, n, cmd);
	return n;
}

struct maccmd
{
	struct cmd;
	uchar action[2];
	uchar addr[6];
};

static int
getmac(Ether *e)
{
	int n;
	struct maccmd cmd;

	memset(&cmd, 0, sizeof(cmd));
	n = mvcmd(e, Cmac, sizeof(cmd)-4, &cmd, sizeof(cmd));
	if(n < 0)
		return -1;
	memcpy(e->addr, cmd.addr, Eaddrlen);
	dprint(2, "%s: mac addr: %.*H\n", argv0, Eaddrlen, e->addr);
	return n;
}

static int
setmac(Ether *e, uchar *mac){
	int n;
	struct maccmd cmd;

	memset(&cmd, 0, sizeof(cmd));
	PUT2(cmd.action, Aset);
	memcpy(cmd.addr, mac, Eaddrlen);
	n = mvcmd(e, Cmac, sizeof(cmd)-4, &cmd, sizeof(cmd));
	if(n < 0)
		return -1;
	memcpy(e->addr, cmd.addr, Eaddrlen);
	dprint(2, "%s: set mac addr to: %.*H\n", argv0, Eaddrlen, e->addr);
	return n;
}

enum
{
	Fwep40	= 0,
	Fwep104	= 0x8000,
	Fwmm	= 0x0800,
	Fprot	= 0x0400,
	Fallmulticast	= 0x0100,
	Fpromiscuous	= 0x0080,
	Fwep	= 0x0008,
	Ftxen	= 0x0002,
	Frxen	= 0x0001,
};

struct macctlcmd
{
	struct cmd;
	uchar flags[4];
};

static int
setmacctl(Ether *e)
{
	int n;
	struct macctlcmd cmd;
	Libertas *liber;

	liber = e->aux;
	memset(&cmd, 0, sizeof(cmd));
	PUT2(cmd.flags, liber->macctl);
	n = mvcmd(e, Cmacctl, sizeof(cmd)-4, &cmd, sizeof(cmd));
	if(n < 0)
		return -1;
	liber->macctl = GET2(cmd.flags);
	return n;
}

static int
mvmulticast(Ether *e, uchar *, int)
{
	Libertas *liber;
	liber = e->aux;
	if(e->nmcasts != 0)
		liber->macctl |= Fallmulticast;
	else
		liber->macctl &= ~Fallmulticast;
	return setmacctl(e);
}

static int
mvpromiscuous(Ether *e, int on)
{
	Libertas *liber;
	liber = e->aux;
	if(on)
		liber->macctl |= Fpromiscuous;
	else
		liber->macctl &= ~Fpromiscuous;
	return setmacctl(e);
}

enum
{
	FC11a	= 1<<10,
	FC11g	= 1<<9,
	FC11b	= 1<<8,
	FCrx0	= 0<<6,
	FCrx1	= 1<<6,
	FCrxboth	= 1<<7,
	FCtx0	= 0<<4,
	FCtx1	= 1<<4,
	FCtxboth	= 1<<5,
	FCexit	= 1<<3,
	FCps	= 1<<1,
	FCwpa = 1<<0,
};

struct hwspeccmd
{
	struct cmd;
	uchar hwifversion[2];
	uchar hwversion[2];
	uchar wcbc[2];
	uchar mcastc[2];
	uchar addr[6];
	uchar region[2];
	uchar antennac[2];
	uchar fwversion[2];
	uchar wcb[4];
	uchar rxpdrdp[4];
	uchar rxpdwrp[4];
	uchar capabilities[4];
};

static int
gethwspec(Ether *e)
{
	int n;
	struct hwspeccmd cmd;
	Libertas *liber;

	liber = e->aux;

	memset(&cmd, 0, sizeof(cmd));
	memset(cmd.addr, 0xff, sizeof(cmd.addr));
	n = mvcmd(e, Cgetspec, sizeof(cmd)-4, &cmd, sizeof(cmd));
	if(n < 0)
		return -1;

	liber->hwversion = GET4(cmd.hwifversion);
	liber->version = GET2(cmd.fwversion);
	liber->flags = GET4(cmd.capabilities);
	liber->region = GET2(cmd.region);
	dprint(2, "%s: gethwspec: hwversion %x version %x flags %x region %d\n",
		argv0, liber->hwversion, liber->version, liber->flags, liber->region);
	
	return 0;
}

enum
{
	/* http://standards.ieee.org/getieee802/download/802.11-2007.pdf */
	Tssid	= 0x00,	/* used by marvell */
	Trates	= 0x01,	/* m */
	Tfhparm	= 0x02,	/* m */
	Tdsparm	= 0x03,	/* m */
	Tcfparm	= 0x04,	/* m */
	Ttim	= 0x05,
	Tibssparm	= 0x06,	/* m */
	Tcountry	= 0x07,	/* m */
	Thopparm	= 0x08,
	Thoppattern	= 0x09,
	Trequest	= 0x0a,
	Tbssload	= 0x0b,
	Tedcaparm	= 0x0c,
	Ttspec	= 0x0d,
	Ttclas	= 0x0e,
	Tchallenge	= 0x10,
	/* [11,1f] resrved */
	Tpowconstrain	= 0x20,	/* m */
	Tpowcapability	= 0x21,	/* m */
	Ttpcreq	= 0x22,
	Ttpcrep	= 0x23,
	Tchan	= 0x24,	/* m */
	Tchanswitch	= 0x25,
	Tmeasreq	= 0x26,
	Tmeasrep	= 0x27,
	Tquiet	= 0x28,	/* m */
	Tibssdfs	= 0x29,	/* m */
	Terpinfo	= 0x2a,
	Ttsdelay	= 0x2b,
	Ttclasproc	= 0x2c,
	/* 2d reserved */
	Tqoscapability	= 0x2e,
	/* 2f reserved */
	Trsn	= 0x30,	/* m */
	/* 31 reserved */
	Texrates	= 0x32,
	/* [33,7e] reserved */
	Texcapabilities	= 0x7f,
	/* [80,dc] reserved */
	Tvendor	= 0xdd,	/* m */
	/* [de,ff] reserved */
	/* these are marvell (libertas?) specific */
	Tkey	= 0x100,
	Tchanparm	= 0x101,
	Tprobe	= 0x102,
	Trssilow	= 0x104,
	Tsnrlow	= 0x105,
	Tfail	= 0x106,
	Tbeac	= 0x107,
	Tledgp	= 0x108,
	Tledb	= 0x109,
	Tpass	= 0x10a,
	Tbprob	= 0x10e,
	Tnssid	= 0x10f,
	Twmm	= 0x110,
	Ttsf	= 0x113,
	Tsleep	= 0x115,
	Trssihigh	= 0x118,
	Tsnrhigh	= 0x119,
};

struct tlv
{
	uchar tag;
	uchar len;
	uchar val[];
};

static int
findtlv(void *buf, int size, int tag, void **r)
{
	struct tlv *t, *end;

	end = (void *)((uchar *)buf + size);
	for(t = buf; t < end; t = (void *)((uchar *)t + t->len + sizeof(*t)))
		if(t->tag == tag){
			*r = t->val;
			return t->len;
		}
	*r = nil;
	return -1;
}

static int
puttlv(void *buf, int size, int tag, int len, void *val)
{
	struct tlv *t;

	if(size < len + sizeof(*t))
		return 0;
	t = buf;
	t->tag = tag;
	t->len = len;
	memmove(t->val, val, len);
	return len + sizeof(*t);
}

struct mvtlv
{
	uchar tag[2];
	uchar len[2];
	uchar val[];
};

static int
findmvtlv(void *buf, int size, int tag, uchar **r)
{
	struct mvtlv *t, *end;

	end = (void *)((uchar *)buf + size);
	for(t = buf; t < end; t = (void *)((uchar *)t + GET2(t->len) + sizeof(*t)))
		if(GET2(t->tag) == tag){
			*r = t->val;
			return GET2(t->len);
		}
	*r = nil;
	return -1;
}

static int
putmvtlv(void *buf, int size, int tag, int len, void *val)
{
	struct mvtlv *t;

	if(size < len + sizeof(*t))
		return 0;
	t = buf;
	PUT2(t->tag, tag);
	PUT2(t->len, len);
	memmove(t->val, val, len);
	return len + sizeof(*t);
}


struct chanparm
{
	uchar type;
	uchar chan;
	uchar scantype;
	uchar mintime[2];
	uchar maxtime[2];
};

struct scancmd
{
	struct cmd;
	union{
	struct{	/* request */
		uchar bsstype;
		uchar bssid[6];
		uchar mvtlv[];
	};
	struct{	/* response */
		uchar bsslen[2];
		uchar apc;
		uchar apbuf[512];
	};
	};
};

struct mvap
{
	uchar len[2];
	uchar bssid[6];
	uchar rssi;
	uchar time[8];
	uchar beacon[2];
	uchar flags[2];
	uchar tlv[];
};

static char *
seprintap(char *s, char *se, void *p)
{
	struct mvap *ap;
	uchar *v;
	char ssid[33]; /* gnar */
	int n, i;

	ap = p;
	s = seprint(s, se, "bssid %.*H rssi %d time %.*H beacon %d flags %x",
		sizeof(ap->bssid), ap->bssid, ap->rssi, sizeof(ap->time), ap->time, GET2(ap->beacon), GET2(ap->flags));

	n = findtlv(ap->tlv, GET2(ap->len)+2-sizeof(*ap), Tssid, &v);
	if(n > 0 && n < 33){
		memcpy(ssid, v, n);
		ssid[n]='\0';
		s = seprint(s, se, " ssid '%s'", ssid);
	}

	n = findtlv(ap->tlv, GET2(ap->len)+2-sizeof(*ap), Trates, &v);
	if(n > 0){
		s = seprint(s, se, " rates");
		for(i=0; i < n; i++)
			s = seprint(s, se, " %d.%d", v[i]/2 & 0x3f, 5*(v[i] & 1));
	}
	n = findtlv(ap->tlv, GET2(ap->len)+2-sizeof(*ap), Texrates, &v);
	if(n > 0)
		for(i=0; i < n; i++)
			s = seprint(s, se, " %d.%d", v[i]/2 & 0x3f, 5*(v[i] & 1));

	n = findtlv(ap->tlv, GET2(ap->len)+2-sizeof(*ap), Tdsparm, &v);
	if(n > 0)
		s = seprint(s, se, " channel %d", v[0]);

	s = seprint(s, se, "\n");

	return s;
}

static struct mvap *
findap(Ether *e, char *ssid, uchar *bssid, int channel){
	int i, rssi, o, n;
	struct mvap *ap, *best;
	Libertas *liber;
	uchar *v;

	liber = e->aux;
	rssi = 256;
	best = nil;
	o = 0;
	for(i = 0; i < liber->scanc; i++, o += GET2(ap->len) + 2){
		ap = (void *)(liber->scanb + o);
		if(bssid && memcmp(ap->bssid, bssid, Eaddrlen)!=0) continue;
		if(ssid){
			n = findtlv(ap->tlv, GET2(ap->len)+2-sizeof(*ap), Tssid, &v);
			if(n<0) continue;
			if(strncmp((char *)v, ssid, n) != 0) continue;
		}
		if(channel){
			n = findtlv(ap->tlv, GET2(ap->len)+2-sizeof(*ap), Tdsparm, &v);
			if(n!=1) continue;
			if(*v != channel) continue;
		}
		if(ap->rssi < rssi){
			best = ap;
			rssi = ap->rssi;
		}
	}
	return best;
}


static int
doscan(Ether *e)
{
	int n, o, i;
	struct scancmd cmd;
	struct chanparm uschannels[11];
	char *bp, *be;
	Libertas *liber;

	liber = e->aux;
	bp = liber->scans;
	be = liber->scans + sizeof(liber->scans);

	memset(&cmd, 0, sizeof(cmd));
	o = sizeof(struct cmd) - 4;
	cmd.bsstype = 3;	/* infratructure or adhoc */
	o++;
//	memset(cmd.bssid, 0, Eaddrlen);
	o+=Eaddrlen;
	
	for(i=0; i < nelem(uschannels); i++){
		uschannels[i].type = 0; /* BG */
		uschannels[i].chan = i+1;
		uschannels[i].scantype = 0;
		PUT2(uschannels[i].mintime, 0);
		PUT2(uschannels[i].maxtime, 256);
	}
	o+=putmvtlv(cmd.mvtlv, sizeof(cmd)-o-4, Tchanparm, sizeof(uschannels), uschannels);
	n = mvcmd(e, Cscan, o, &cmd, sizeof(cmd));
	if(n < 0)
		return n;

	liber->scanc = cmd.apc;
	liber->scanlen = GET2(cmd.bsslen);
	memmove(liber->scanb, cmd.apbuf, liber->scanlen);
	o = 0;
	for(i = 0; i < liber->scanc; i++){
		bp = seprintap(bp, be, cmd.apbuf + o);
		o += GET2(cmd.apbuf + o) + 2;
	}
	return liber->scanc;
}

static char *
seprintscan(char *s, char *se, Ether *e){
	Libertas *liber;
	liber = e->aux;
	return seprint(s, se, "%s", liber->scans);
}

struct associatecmd {
	struct cmd;
	uchar addr[6];
	uchar flags[2];
	uchar listen[2];
	uchar interval[2];
	uchar dtim;
	uchar data[15+5+4+12];
};

static int
doassociate(Ether *e, char *ssid, uchar *bssid, int chan)
{
	int n, o, ssidlen;
	struct associatecmd cmd;
	uchar tmp, *ch;
	struct mvap *ap;
	Libertas *liber;

	liber = e->aux;

	if(bssid && memcmp(bssid, "\0\0\0\0\0\0", Eaddrlen) == 0)
		bssid = nil;
	if(ssid && strlen(ssid) == 0)
		ssid = nil;

	SET(ssidlen);
	if(ssid)
		ssidlen = strlen(ssid);

	if(ssid==nil || bssid==nil || chan==0){
		ap = findap(e, ssid, bssid, chan);
		if(ap == nil)
			return -1;
		if(ssid == nil)
			ssidlen = findtlv(ap->tlv, GET2(ap->len)-sizeof(*ap), Tssid, &ssid);
		if(chan == 0){
			findtlv(ap->tlv, GET2(ap->len)-sizeof(*ap), Tdsparm, &ch);
			chan = *ch;
		}
		if(bssid == nil)
			bssid = ap->bssid;
	}

	memset(&cmd, 0, sizeof(cmd));
	o = sizeof(struct cmd)-4;
	if(bssid)
		memcpy(cmd.addr, bssid, Eaddrlen);
	o += Eaddrlen;
	PUT2(cmd.listen, 10); /* apparently */
	PUT2(cmd.flags, 0x401);
	/* ignore interval and dtim */
	o += 2 + 2 + 2 + 1;

	if(ssid)
		o += putmvtlv(cmd.cmd+o, sizeof(cmd)-o-4, Tssid, ssidlen, ssid);
	if(chan){
		tmp = chan;
		o += putmvtlv(cmd.cmd+o, sizeof(cmd)-o-4, Tdsparm, 1, &tmp);
	}
	o += putmvtlv(cmd.cmd+o, sizeof(cmd)-o-4, Tcfparm, 0, nil);
	o += putmvtlv(cmd.cmd+o, sizeof(cmd)-o-4, Trates, 8, "\x82\x84\x8b\x96\xa4\xb0\xc6\xec");

	n = mvcmd(e, Cassociate, o, &cmd, sizeof(cmd));

	if(GET2(cmd.result)){
		strncpy(liber->ssid, ssid, ssidlen);
		memcpy(liber->bssid, bssid, Eaddrlen);
		liber->channel = chan;
		return 0;
	}
	if(n < 0)
		return -1;
	return n;
}


static int
finalinit(Ether *e)
{
	Libertas *liber;
	liber = e->aux;

	getmac(e);
	gethwspec(e);
	liber->macctl = Ftxen | Frxen;
	setmacctl(e);
	doscan(e);
	dprint(2, "%s", liber->scans);
	liber->try_ssid[0] = 0;
	memcpy(liber->try_bssid, "\0\0\0\0\0\0", Eaddrlen);
	liber->channel = 0;
	doassociate(e, nil, nil, 0);
	return 1;
}

static int
ctlrinit(Ether *e)
{
	if(e->cid == MV8388){
		dprint(2, "%s: found the MV8388\n", argv0);
		e->init2 = mvfirmware;
		e->init3 = finalinit;
		return 0;
	}
	return -1;
}

static void
mvfree(Ether *ether)
{
	dprint(2, "%s: mvfree %#p\n", argv0, ether);
	free(ether->aux);
	ether->aux = nil;
}

long
mvbread(Ether *e, Buf *bp)
{
	Libertas *liber;

	liber = e->aux;
	bp->rp = bp->data + Hdrsize;

	while(1){
		bp->ndata = -1;
		bp->ndata = read(e->epin->dfd, bp->rp-4, sizeof(bp->data)-Hdrsize+4);
		if(bp->ndata < 0){
			dprint(2, "%s: mvbread: %r\n", argv0);
			return bp->ndata;
		}

		ddprint(2, "%s: mvbread: got %d bytes: %.*H\n", argv0, bp->ndata, bp->ndata, bp->rp-4);
		switch(GET4(bp->rp-4)){
		case Mdata:
			bp->rp += 28;	/* wee */
			memmove(bp->rp, bp->rp-8, 12);
			return bp->ndata;
			break;
		case Mrequest:
			if(liber->cmd == nil) break;
			if(GET2((uchar*)liber->cmd+8) != GET2(bp->rp+4))	/* sequence nums */
				break;
			qlock(&liber->l);
			if(bp->ndata <= liber->respsz){
				memmove(liber->cmd, bp->rp-4, bp->ndata);
				liber->respsz = bp->ndata;
			}else
				liber->respsz = -bp->ndata;
			liber->cmdresp = 1;
			rwakeup(&liber->vous);
			qunlock(&liber->l);
			break;
		}
	}
}

struct txpd
{
	uchar status[4];
	uchar control[4];
	uchar pktoff[4];
	uchar pktlen[2];
	uchar dst[6];
	uchar priority;
	uchar flags;
	uchar pad[2];
};

struct etherframe
{
	uchar dst[6];
	uchar src[6];
	uchar type[2];
};

static long
mvbwrite(Ether *e, Buf *bp)
{
	long n;
	struct txpd *txpd;
	struct etherframe *ether;

	dprint(2, "%s: etherbwrite %d bytes\n", argv0, bp->ndata);
	ether = (void *)bp->rp;
	txpd = (void *)(bp->rp - sizeof(*txpd));
	PUT4(bp->rp - 4 - sizeof(*txpd), Mdata);

	memset(txpd, 0, sizeof(*txpd));
	PUT4(txpd->pktoff, sizeof(*txpd));
	PUT2(txpd->pktlen, bp->ndata);
	memcpy(txpd->dst, ether->dst, Eaddrlen);
	
	n = write(e->epout->dfd, bp->rp - 4 - sizeof(*txpd), bp->ndata + 4 + sizeof(*txpd));
	if(n < 0){
		dprint(2, "%s: etherbwrite: %r\n", argv0);	/* keep { and }  */
	}else
		dprint(2, "%s: etherbwrite wrote %ld bytes\n", argv0, n);
	if(n <= 0)
		return n;
	if((bp->ndata % e->epout->maxpkt) == 0){
		dprint(2, "%s: short pkt write\n", argv0);
		write(e->epout->dfd, "", 1);
	}
	return n;
}

int
mvctl(Ether *e, char *s){
	Libertas *liber;

	liber = e->aux;
	if(strncmp(s, "scan", 4) == 0){
		doscan(e);
		return 1;
	}
	if(strncmp(s, "associate", 9) == 0){
		doassociate(e, liber->try_ssid, liber->try_bssid, liber->channel);
		return 1;
	}
	if(strncmp(s, "ssid", 4) == 0){
		if(s[strlen(s)-1] == '\n')
			s[strlen(s)-1] = 0;
		strcpy(liber->try_ssid, s + 5);
		return 1;
	}
	if(strncmp(s, "bssid", 5) == 0){
		if(strlen(s + 6) >= 12)
			dec16(liber->try_bssid, Eaddrlen, s + 6, Eaddrlen*2);
		else
			memcpy(liber->try_bssid, "\0\0\0\0\0\0", Eaddrlen);
		return 1;
	}
	if(strncmp(s, "channel", 7) == 0){
		liber->try_channel = atoi(s+8);
		return 1;
	}
	if(strncmp(s, "reset", 5) == 0){
		mvreset(e);
		threadexitsall("");
		return 1;
	}
	return -1;
}
		

int
mv8388reset(Ether *e)
{
	Cinfo *ip;
	Dev *dev;
	Libertas *liber;

	fmtinstall('H', encodefmt);
	dev = e->dev;
	for(ip = winfo; ip->vid != 0; ip++)
		if(ip->vid == dev->usb->vid && ip->did == dev->usb->did){
			e->cid = ip->cid;
			if(ctlrinit(e) < 0){
				dprint(2, "%s: init failed: %r\n", argv0);
				return -1;
			}
			e->aux = emallocz(sizeof(Libertas), 1);
			liber = e->aux;
			liber->vous.l = &liber->l;
			e->bread = mvbread;
			e->bwrite = mvbwrite;
			e->free = mvfree;
			e->promiscuous = mvpromiscuous;
			e->multicast = mvmulticast;
			e->seprintstats = seprintscan;
			e->ctl = mvctl;
			e->mbps = 27;	/* this would require some creativity */
			return 0;
		}
	return -1;
}